
#include <iostream>
#include <string>
#include <z3++.h>
#include <cmath>
#include <fstream>
#include <getopt.h>

#include <chrono>

#include "UnconstrainedVariableSimplifier.h"
#include "ExprSimplifier.h"

using namespace std;
using namespace z3;

using std::chrono::high_resolution_clock;
using std::chrono::milliseconds;

int main(int argc, char* argv[])
{
    static struct option long_options[] = {
	{"propagate-unconstrained", no_argument, 0, 'p' },
	{"count-locally", no_argument, 0, 'l' },
	{"canonize", no_argument, 0, 'c' },
	{"run-z3", no_argument, 0, 'z' },
	{"inprocess", no_argument, 0, 'i' },
	{"full", no_argument, 0, 'f' },
	{0,           0,                 0,  0   }
    };

    bool propagateUnconstrainedFlag = false, runZ3 = false, canonize = false, countVariablesLocally = false, inprocess = false, full = false;
    char* filename;

    int opt = 0;

    int long_index = 0;
    while ((opt = getopt_long(argc, argv,"plzcif", long_options, &long_index )) != -1)
    {
	switch (opt)
	{
	case 'p':
	    propagateUnconstrainedFlag = true;
	    break;
	case 'l':
	    countVariablesLocally = true;
	    break;
	case 'z':
	    runZ3 = true;
	    break;
	case 'i':
	    inprocess = true;
	    break;
	case 'c':
	    canonize = true;
	    break;
	case 'f':
	    full = true;
	    break;
	default:
	    std::cout << "Invalid arguments" << std::endl;
	    exit(1);
	}
    }

    if (optind < argc)
    {
	filename = argv[optind];
    }
    else
    {
	std::cout << "Filename required" << std::endl;
	return 1;
    }

    z3::context ctx;
    Z3_ast ast = Z3_parse_smtlib2_file(ctx, filename, 0, 0, 0, 0, 0, 0);
    expr e = to_expr(ctx, ast);

    e = e.simplify();

    if (canonize)
    {
	UnconstrainedVariableSimplifier simplifier(ctx, e);
	e = simplifier.CanonizeBoundVariables(e);
    }

    if (full)
    {
	ExprSimplifier simplifier(ctx, propagateUnconstrainedFlag);
	e = simplifier.Simplify(e);
    }
    else if (propagateUnconstrainedFlag)
    {
	unsigned oldHash = 0;

	do
	{
	    oldHash = e.hash();
	    UnconstrainedVariableSimplifier simplifier(ctx, e);
	    simplifier.SetCountVariablesLocally(countVariablesLocally);
	    //simplifier.PrintUnconstrained();
	    simplifier.SimplifyIte();
	    e = simplifier.GetExpr();

	    if (inprocess)
	    {
		z3::goal g(ctx);
		g.add(e);

		z3::tactic derTactic =
		    z3::tactic(ctx, "simplify") &
		    z3::tactic(ctx, "elim-and") &
		    z3::tactic(ctx, "der") &
		    z3::tactic(ctx, "simplify");

		z3::apply_result result = derTactic(g);

		assert(result.size() == 1);
		z3::goal simplified = result[0];
		e = simplified.as_expr();
	    }
	} while (oldHash != e.hash());
    }

    e = e.simplify();
    //std::cout << e << std::endl;

    if (runZ3)
    {
        solver s(ctx);
	params p(ctx);

	//	p.set(":max_memory", 3u);
	s.set(p);

	s.add(e);
	std::cout << s.check() << std::endl;
    }
    else
    {
	std::cout << e << std::endl;
    }
}
